@@ -59,6 +59,7 @@ enum {
 	REQUEST_OBJECT,
 	ON_URL,
 	ACTIVATE,
+	MOVE_FOCUS_OUT,
 	LAST_SIGNAL
 };
 
@@ -158,23 +159,31 @@ html_view_set_adjustments (HtmlView *view)
 	GtkAdjustment *vertical, *horizontal;
 
 	if (view->root) {
+		gint focus_width;
+		gint focus_pad;
+
+		gtk_widget_style_get (GTK_WIDGET (view),
+				      "focus-line-width", &focus_width,
+				      "focus-padding", &focus_pad,
+				      NULL);
+
 		layout = GTK_LAYOUT (view);
 		vertical = layout->vadjustment;
 		horizontal = layout->hadjustment;
 		
 		vertical->lower = 0;
-		vertical->upper = view->root->height;
+		vertical->upper = view->root->height + 2 * (focus_width + 2 * focus_pad);
 		vertical->page_size = (gfloat)GTK_WIDGET (view)->allocation.height;
 		vertical->step_increment = GTK_WIDGET (view)->allocation.height / 10.0;
 		vertical->page_increment = GTK_WIDGET (view)->allocation.height * 0.9;
 
 		horizontal->lower = 0.0;
-		horizontal->upper = view->root->width;
+		horizontal->upper = view->root->width + 2 * (focus_width + 2 * focus_pad);
 		horizontal->page_size = GTK_WIDGET (view)->allocation.width;
 		horizontal->step_increment = GTK_WIDGET (view)->allocation.width / 10.0;
 		horizontal->page_increment = GTK_WIDGET (view)->allocation.width * 0.9;
 
-		gtk_layout_set_size (layout, view->root->width, view->root->height);
+		gtk_layout_set_size (layout, horizontal->upper, vertical->upper);
 		
 		gtk_adjustment_changed (vertical);
 		gtk_adjustment_changed (horizontal);
@@ -187,12 +196,19 @@ html_view_relayout (HtmlView *view)
 	/* Tell the root layout box to relayout itself */
 	if (view->painter && view->root) {
 		HtmlRelayout *relayout;
+		gint focus_width;
+		gint focus_pad;
+
+		gtk_widget_style_get (GTK_WIDGET (view),
+				      "focus-line-width", &focus_width,
+				      "focus-padding", &focus_pad,
+				      NULL);
+
+		view->root->x = focus_width + 2 * focus_pad;
+		view->root->y = focus_width + 2 * focus_pad;
 		
-		view->root->x = 0;
-		view->root->y = 0;
-		
-		HTML_BOX_ROOT (view->root)->min_width = GTK_WIDGET (view)->allocation.width;
-		HTML_BOX_ROOT (view->root)->min_height = GTK_WIDGET (view)->allocation.height;
+		HTML_BOX_ROOT (view->root)->min_width = GTK_WIDGET (view)->allocation.width - 2 * (focus_width + 2 *focus_pad);
+		HTML_BOX_ROOT (view->root)->min_height = GTK_WIDGET (view)->allocation.height - 2 * (focus_width + 2 *focus_pad);
 		
 		relayout = html_relayout_new ();
 		relayout->type = HTML_RELAYOUT_INCREMENTAL;
@@ -633,6 +649,25 @@ html_view_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 	html_view_relayout (view);
 }
 
+static void
+html_view_size_request (GtkWidget *widget, GtkRequisition *requisition)
+{
+	HtmlView *view = HTML_VIEW (widget);
+	gint focus_width;
+	gint focus_pad;
+
+	if (GTK_WIDGET_CLASS (parent_class)->size_request)
+		( *GTK_WIDGET_CLASS (parent_class)->size_request) (widget, requisition);
+
+	gtk_widget_style_get (widget,
+			      "focus-line-width", &focus_width,
+			      "focus-padding", &focus_pad,
+			      NULL);
+
+	requisition->width += 2 * (focus_width + 2 * focus_pad);
+	requisition->height += 2 * (focus_width + 2 * focus_pad);
+}
+
 static void
 html_view_realize (GtkWidget *widget)
 {
@@ -681,6 +716,7 @@ static gboolean
 html_view_expose (GtkWidget *widget, GdkEventExpose *event)
 {
 	HtmlView *view;
+	gboolean retval;
 
 	view = HTML_VIEW (widget);
 
@@ -688,8 +724,33 @@ html_view_expose (GtkWidget *widget, GdkEventExpose *event)
 		return FALSE;
 
 	html_view_paint (view, &event->area);
+	retval = GTK_WIDGET_CLASS (parent_class)->expose_event (widget, event);
+	if (GTK_WIDGET_HAS_FOCUS (widget) &&
+	    !view->document->focus_element) {
+		gint focus_width;
+		gint focus_pad;
+		gint x;
+		gint y;
+		gint width;
+		gint height;
 
-	return GTK_WIDGET_CLASS (parent_class)->expose_event (widget, event);
+		gtk_widget_style_get (widget,
+				      "focus-line-width", &focus_width,
+				      "focus-padding", &focus_pad,
+				      NULL);
+		x = focus_pad;
+		y = focus_pad;
+		width = widget->allocation.width - (focus_width + 2 * focus_pad);
+		height = widget->allocation.height - (focus_width + 2 * focus_pad);
+		gtk_paint_focus (widget->style, GTK_LAYOUT (view)->bin_window,
+				 GTK_WIDGET_STATE (widget),
+				 &event->area, widget, "htmlview",
+				 x,
+				 y,
+				 width,
+				 height);
+	}
+	return retval;
 }
 
 
@@ -740,15 +801,13 @@ html_view_focus_in (GtkWidget *widget, GdkEventFocus *event)
 	HtmlView *view;
 	DomElement *element;
 
-	GTK_WIDGET_SET_FLAGS (widget, GTK_HAS_FOCUS);
-
 	view = HTML_VIEW (widget);
 
 	element = html_view_get_and_unset_saved_focus (view);
 	if (element) {
 		html_document_update_focus_element (view->document, element);
 	}
-	return FALSE;
+	return GTK_WIDGET_CLASS (parent_class)->focus_in_event (widget, event);
 }
 
 static gboolean
@@ -758,15 +817,12 @@ html_view_focus_out (GtkWidget *widget, GdkEventFocus *event)
 
 	view = HTML_VIEW (widget);
 
-	GTK_WIDGET_UNSET_FLAGS (widget, GTK_HAS_FOCUS);
-		
 	if (view->document->focus_element) {
 		html_view_set_saved_focus (view);
 		if (GTK_CONTAINER (widget)->focus_child == NULL)
 			html_document_update_focus_element (view->document, NULL);
 	}
-
-	return FALSE;
+	return GTK_WIDGET_CLASS (parent_class)->focus_out_event (widget, event);
 }
 
 #if 0
@@ -807,32 +863,29 @@ html_view_focus (GtkWidget *widget, GtkDirectionType direction)
 {
 	HtmlView *view = HTML_VIEW (widget);
 	HtmlBox *box;
-	DomElement *old_focus_element = view->document->focus_element;
+	DomElement *old_focus_element;
 
-	if (view->document == NULL || view->document->dom_document == NULL) {
-		if (!GTK_WIDGET_HAS_FOCUS (widget)) {
-			gtk_widget_grab_focus (widget);
-			return TRUE;
-		}
+	if (view->moving_focus_out)
+		return FALSE;
+
+	if (!GTK_WIDGET_HAS_FOCUS (widget)) {
+		gtk_widget_grab_focus (widget);
+		return TRUE;
+	} else if (view->document == NULL || view->document->dom_document == NULL) {
 		return FALSE;
 	}
 
-
+	old_focus_element = view->document->focus_element;
 	if (direction == GTK_DIR_TAB_FORWARD) {
 		html_document_update_focus_element (view->document, html_focus_iterator_next_element (view->document->dom_document, view->document->focus_element));
-		if (view->document->focus_element) {
-			html_view_focus_element (view);
-			return TRUE;
-		}
+		html_view_focus_element (view);
+		return TRUE;
 	}
 	else if (direction == GTK_DIR_TAB_BACKWARD) {
 		html_document_update_focus_element (view->document, html_focus_iterator_prev_element (view->document->dom_document, view->document->focus_element));
 
-		if (view->document->focus_element) {
-			html_view_focus_element (view);
-			return TRUE;
-		}
-
+		html_view_focus_element (view);
+		return TRUE;
 	}
 
 	/* Handle special case when we haven't got any focusable elements */
@@ -903,6 +956,24 @@ html_view_real_activate (HtmlView *view)
 	html_event_activate (view);
 }
 
+static void html_view_real_move_focus_out (HtmlView *view, GtkDirectionType direction_type)
+{
+	GtkWidget *toplevel;
+
+	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (view));
+	if (!GTK_WIDGET_TOPLEVEL (toplevel))
+		return;
+
+	html_document_update_focus_element (view->document, NULL);
+	g_object_ref (view);
+
+	view->moving_focus_out = TRUE;
+	g_signal_emit_by_name (toplevel, "move_focus", direction_type);
+	view->moving_focus_out = FALSE;
+
+	g_object_unref (view);
+}
+
 static void
 html_view_add_move_binding (GtkBindingSet *binding_set, guint keyval, guint modmask, GtkMovementStep step, gint count)
 {
@@ -911,7 +982,17 @@ html_view_add_move_binding (GtkBindingSet *binding_set, guint keyval, guint modm
 				      GTK_TYPE_ENUM, step,
 				      GTK_TYPE_INT, count,
 				      GTK_TYPE_BOOL, FALSE);
+}
 
+static void
+html_view_add_tab_binding (GtkBindingSet *binding_set, GdkModifierType modifiers, GtkDirectionType direction)
+{
+	gtk_binding_entry_add_signal (binding_set, GDK_Tab, modifiers,
+				      "move_focus_out", 1,
+				      GTK_TYPE_DIRECTION_TYPE, direction);
+	gtk_binding_entry_add_signal (binding_set, GDK_KP_Tab, modifiers,
+				      "move_focus_out", 1,
+				      GTK_TYPE_DIRECTION_TYPE, direction);
 }
 
 void
@@ -1059,6 +1140,7 @@ html_view_class_init (HtmlViewClass *klass)
 	widget_class->focus = html_view_focus;
 	
 	widget_class->size_allocate = html_view_size_allocate;
+	widget_class->size_request = html_view_size_request;
 	widget_class->realize = html_view_realize;
 	widget_class->expose_event = html_view_expose;
 	widget_class->motion_notify_event = html_view_motion_notify;
@@ -1079,6 +1161,7 @@ html_view_class_init (HtmlViewClass *klass)
 	
 	klass->move_cursor = html_view_real_move_cursor;
 	klass->activate = html_view_real_activate;
+	klass->move_focus_out = html_view_real_move_focus_out;
 	
 	view_signals [MOVE_CURSOR] =
 		g_signal_new ("move_cursor",
@@ -1114,13 +1197,24 @@ html_view_class_init (HtmlViewClass *klass)
 			      G_TYPE_STRING);
 
 	view_signals [ACTIVATE] =
-		g_signal_new ("ACTIVATE",
+		g_signal_new ("activate",
 			      G_TYPE_FROM_CLASS (object_class),
 			      G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
 			      G_STRUCT_OFFSET (HtmlViewClass, activate),
 			      NULL, NULL,
 			      html_marshal_VOID__VOID,
 			      G_TYPE_NONE, 0);
+
+	view_signals [MOVE_FOCUS_OUT] =
+		g_signal_new ("move_focus_out",
+			      G_TYPE_FROM_CLASS (object_class),
+			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
+			      G_STRUCT_OFFSET (HtmlViewClass, move_focus_out),
+			      NULL, NULL,
+			      html_marshal_VOID__ENUM,
+			      G_TYPE_NONE, 1,
+			      GTK_TYPE_DIRECTION_TYPE);
+
 	widget_class->activate_signal = view_signals[ACTIVATE];
 
 	html_view_add_move_binding (binding_set, GDK_Page_Down, 0,
@@ -1142,6 +1236,8 @@ html_view_class_init (HtmlViewClass *klass)
 	html_view_add_move_binding (binding_set, GDK_Left, 0,
 				    GTK_MOVEMENT_VISUAL_POSITIONS, -1);
 	
+	html_view_add_tab_binding (binding_set, GDK_CONTROL_MASK, GTK_DIR_TAB_FORWARD);
+	html_view_add_tab_binding (binding_set, GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_DIR_TAB_BACKWARD);
 	
 }
 
@@ -1162,6 +1258,7 @@ html_view_init (HtmlView *view)
 	view->relayout_idle_id = 0;
 	view->relayout_timeout_id = 0;
 	view->magnification = 1.0;
+	view->moving_focus_out = FALSE;
 }
 
 GType
@@ -1247,14 +1344,21 @@ html_view_focus_element (HtmlView *view)
 {
 	HtmlBox *box;
 
-	html_view_scroll_to_node (view, DOM_NODE (view->document->focus_element), HTML_VIEW_SCROLL_TO_BOTTOM);
+	if (view->document->focus_element) {
+		html_view_scroll_to_node (view, DOM_NODE (view->document->focus_element), HTML_VIEW_SCROLL_TO_BOTTOM);
 
-	box = html_view_find_layout_box (view, DOM_NODE (view->document->focus_element), FALSE);
-	if (box && HTML_IS_BOX_EMBEDDED (box)) 
-		gtk_widget_child_focus (HTML_BOX_EMBEDDED (box)->widget, GTK_DIR_TAB_FORWARD);
-	else
-		gtk_widget_grab_focus (GTK_WIDGET (view));
+		box = html_view_find_layout_box (view, DOM_NODE (view->document->focus_element), FALSE);
+		if (box && HTML_IS_BOX_EMBEDDED (box)) 
+			gtk_widget_child_focus (HTML_BOX_EMBEDDED (box)->widget, GTK_DIR_TAB_FORWARD);
+		else
+			gtk_widget_grab_focus (GTK_WIDGET (view));
+	} else {
+		/* No element focused to scroll to top */
+		GtkAdjustment *adj = GTK_LAYOUT (view)->vadjustment;
 
+		set_adjustment_clamped (adj, 0);
+	}
+	gtk_widget_queue_draw (GTK_WIDGET (view));
 }
 
 #ifdef ENABLE_ACCESSIBILITY
